"""bsw file management functions"""

import shutil
import os


class FileManager(object):
    def __init__(self, out_dir):
        self.out_dir = out_dir

    def create_out_dir(self):
        """Create the output dir if it doens't exist."""
        if not os.path.isdir(self.out_dir):
            os.makedirs(self.out_dir)

    def merge_dirs(self, source_path, dest_path):
        """Copy the contents of source_path to dest_path, creating subdirectories where
        neccessary.

        Will not overwrite contents in dest_path.
        """
        source_path_abs = os.path.abspath(source_path)
        for (dirpath, dirnames, filenames) in os.walk(source_path_abs):
            for filename in filenames:
                source_path_file = os.path.join(dirpath, filename)
                source_path_rel = source_path_file.replace(source_path_abs,
                    "", 1).lstrip(os.path.sep)
                source_path_dir = os.path.dirname(source_path_rel)
                if not os.path.isdir(os.path.join(dest_path, source_path_dir)):
                    os.makedirs(os.path.join(dest_path, source_path_dir))
                if not os.path.isfile(os.path.join(dest_path,
                                                   source_path_rel)):
                    shutil.copyfile(source_path_file, os.path.join(dest_path,
                        source_path_rel))

    def copy_template_assets(self):
        """Copy template asset files to output directory."""
        if os.path.isdir(os.path.join("templates", "static")):
            self.merge_dirs(os.path.join("templates", "static"),
                os.path.join(self.out_dir, "static"))

    def copy_site_assets(self):
        """Copy site asset files to output directory."""
        if os.path.isdir(os.path.join(self.out_dir, "static")):
            self.merge_dirs("static", os.path.join(self.out_dir, "static"))

    def clean_build_path(self):
        """Remove build folder if it exists."""
        if os.path.isdir(self.out_dir):
            shutil.rmtree(self.out_dir)

    def check_required_paths(self):
        """Check that the required bsw project paths exist"""
        template_path = os.path.join(".", "templates", "base.html")
        if not os.path.exists(template_path):
            raise IOError("Base template (templates/base.html) "
                                    "not found")

    def init_with_template(self, template_name, destination):
        """Initialise the destination folder with the specified template
        skeleton.
        """
        module_path = os.path.dirname(os.path.abspath(__file__))
        template_path = os.path.join(module_path, "templates", template_name)
        
        if not os.path.exists(template_path):
            raise IOError("Template {0} not found".format(template_name))

        self.merge_dirs(template_path, destination)



"""bsw page model and functions"""

import os
import re

from . import templates


class Page:
    def __init__(self, filename):
        self.filename = filename
        self.body = None  # The page body after var replacement
        self.body_raw = None  # The un-transformed page body
        self.rendered_page = None  # Finished page rendered with template
        self.page_vars = {}

    def load_and_parse(self):
        self.load()
        self.extract_vars()
        self.strip_vars()

    def load(self):
        """Load page from file."""
        with open(os.path.join("pages", self.filename), "r") as page_file:
            self.body_raw = page_file.read()

    def extract_vars(self):
        """Populate self.vars with vars extracted from page."""
        regex_var_capture = "<!--\s+(\w+)\s+=\s+\"([^>]*)\"\s+-->"
        matches = re.findall(regex_var_capture, self.body_raw)
        for match in matches:
            self.page_vars[match[0]] = match[1]

    def strip_vars(self):
        """Strip page vars from raw body and populate self.body."""
        page_data = self.body_raw
        for page_var in self.page_vars:
            regex_this_var = "<!--\s+({0})\s+=\s+\"({1})\"\s+-->".format(
                                page_var,
                                self.page_vars[page_var])
            page_data = re.sub(regex_this_var, "", page_data)
        self.body = page_data

    def replace_includes(self):
        """Replace <!-- include("my_include.html") --> directives with the
        referenced file.
        """
        regex_include = "<!--\s+include\(\"([^>]*)\"\)\s+-->"
        matches = re.findall(regex_include, self.body)
        for match in matches:
            include_data = includes.get_include(match)
            regex_this_include = "<!--\s+include\(\"{0}\"\)\s+-->".format(match)
            self.rendered_page = re.sub(regex_this_include,
                                        include_data,
                                        self.rendered_page)

    def render(self):
        """Render page using template, insert any includes and do page
        var replacement.
        """
        if "template" in self.page_vars:
            template = templates.get_template(self.page_vars["template"])
        else:
            template = templates.get_template("base.html")

        self.rendered_page = template.replace("$page_content", self.body)
        self.replace_includes()
        for page_var in self.page_vars:
            self.rendered_page = self.rendered_page.replace("$" + page_var,
                                                            self.page_vars[page_var])


def collect_pages():
    """Collect source page from "pages" folder in pwd"""
    page_paths = []
    pages = []

    for (dirpath, dirnames, filenames) in os.walk("pages/"):
        for filename in filenames:
            if filename.endswith(".html") or filename.endswith(".htm"):
                page_paths.append(
                    os.path.join(dirpath, filename).split(
                                    os.path.sep, 1)[1])

    for page_path in page_paths:
        new_page = Page(page_path)
        pages.append(new_page)

    return pages


def write_pages(pages, out_dir):
    """Write rendered page to out_dir"""
    for page in pages:
        if not os.path.isdir(
                os.path.join(out_dir, os.path.dirname(page.filename))):
            os.makedirs(os.path.join(out_dir, os.path.dirname(page.filename)))
        with open(os.path.join(out_dir, page.filename), "w") as out_file:
            out_file.write(page.rendered_page)



"""bsw - Build Static Website

A simple static website generator operating on the current directory and
building static websites from pages and templates.

Your project folder should be structured as follows:
    pages/
    templates/
        base.html

An example layout would be:
    pages/
        about/
            index.html
        index.html
    templates/
        static/
            site_logo.png
        includes/
            social_media_links.html
        base.html
        blog.html
    static/
        images/
            blog_header_20160101.png


This example has a base template (base.html), a blog template (blog.html),
an includes file with social media links (social_media_links.html),
and default site page (index.html) and an about page (about/index.html).

Static files from both the site statics (static/) and template statics
(templates/static/) folders are combined into the build/ folder
(automatically created when the site is first generated).

The structure in pages/ is preserved and cloned to the build/ folder.
"""

import argparse
import os

try:
    # Python 3
    import http.server
    import socketserver
except ImportError:
    # Python 2
    import SimpleHTTPServer
    import SocketServer

import sys

from . import files
from . import pages
from . import templates

OUT_DIR = os.path.abspath(os.path.join(".", "build"))


def serve_content():
    """Serve rendered content on SimpleHTTPServer. This helps with absolute
    file paths such as /static/css/main.css, which wouldn't work
    properly if opened directly in from file:///"""
    PORT = 8000
    os.chdir(OUT_DIR)

    # Allow us to quickly kill and restart server without waiting for TCP
    # socket to close down completely
    try:
        # Python 3
        handler = http.server.SimpleHTTPRequestHandler
        socketserver.TCPServer.allow_reuse_address = 1
        httpd = socketserver.TCPServer(("", PORT), handler)
    except NameError:
        # Python 2
        handler = SimpleHTTPServer.SimpleHTTPRequestHandler
        SocketServer.TCPServer.allow_reuse_address = 1
        httpd = SocketServer.TCPServer(("", PORT), handler)

    print("Serving content at localhost: " + str(PORT))
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("Exiting")
        sys.exit()


def build_static_web(clean_build_path):
    file_manager = files.FileManager(OUT_DIR)

    # Check that required paths (e.g. base template) exist before starting
    try:
        file_manager.check_required_paths()
    except IOError as ex:
        print("Error: {0}".format(ex))
        print("Did you remember to initialise the site with --init?")
        sys.exit(1)

    if (clean_build_path):
        print("Cleaning build path...")
        file_manager.clean_build_path()
    file_manager.create_out_dir()

    print("Loading pages...")
    site_pages = pages.collect_pages()
    for page in site_pages:
        page.load_and_parse()

    # Check that any explicit templates exist before rendering
    try:
        templates.check_templates_exist(site_pages)
    except IOError as ex:
        print("Error: {0}".format(ex))
        sys.exit(1)

    print("Rendering pages...")
    for page in site_pages:
        page.render()
    print("Writing pages...")
    pages.write_pages(site_pages, OUT_DIR)

    print("Copying site and template assets...")
    file_manager.copy_template_assets()
    file_manager.copy_site_assets()

    print("Static site build complete.")


def main():
    parser = argparse.ArgumentParser(
        description="bsw - build static website")
    parser.add_argument("-C", "--clean", action="store_true",
                        help="remove existing build folder before building")
    parser.add_argument("-s", "--serve", action="store_true",
                        help="serve content after build (default port 8000)")
    parser.add_argument("--init", 
                        help="initialise a new bsw site as the specified path",
                        nargs="?",
                        const=".",
                        metavar="PATH")
    parser.add_argument("--template", 
                        help="built-in template to use with --init")
    args = parser.parse_args()

    if (args.clean and args.init):
        print("Error: Please specify either --clean or --init")
        sys.exit(1)

    if (args.serve and args.init):
        print("Error: Please specify either --serve or --init")
        sys.exit(1)

    if (args.template and not args.init):
        print("Error: --template can only be used with --init")
        sys.exit(1)

    if args.init:
        file_manager = files.FileManager(OUT_DIR)
        template = args.template if args.template else "default"
        file_manager.init_with_template(template, args.init)
        print("bsw site initialised")
        sys.exit(0)

    clean_build_path = False
    if args.clean:
        clean_build_path = True

    build_static_web(clean_build_path)

    if args.serve:
        serve_content()



"""bsw template handling functions"""

import os

template_cache = {}


def get_template(path):
    """Return template body (cached)"""
    if path not in template_cache:
        with open(os.path.join(".", "templates", path), "r") as template_file:
            template_cache[path] = template_file.read()
    return template_cache[path]


def check_templates_exist(pages):
    """Checks that all templates exist, raises FileNotFoundError if missing"""
    for page in pages:
        if "template" in page.page_vars:
            template_path = os.path.join(".", "templates",
                                page.page_vars["template"])
            if not os.path.exists(template_path):
                raise FileNotFoundError(
                    "Template '{0}' for page {1} not found".format(
                        template_path,
                        page.filename))



"""setup.py: setuptools control.

Borrowed from:
    https://gehrcke.de/2014/02/distributing-a-python-command-line-application/
"""


import re
from setuptools import setup


version = re.search(
    "^__version__\s*=\s*\"(.*)\"",
    open("bsw/__init__.py").read(),
    re.M
).group(1)


with open("README.rst", "rb") as f:
    long_descr = f.read().decode("utf-8")


setup(
    name = "bsw",
    packages = ["bsw"],
    entry_points = {
        "console_scripts": ["bsw = bsw.bsw:main"]
    },
    version = version,
    description = "bsw - Build Static Web, a simple static website builder.",
    long_description = long_descr,
    author = "Dave Barker",
    author_email = "david@dbark.co.uk",
    url = "https://github.com/davb5/bsw",
)



"""bsw include handling functions"""

include_cache = {}


def get_include(path):
    """Get include page contents (cached)"""
    if path not in include_cache:
        full_path = os.path.join("templates", "includes", path)

        with open(full_path, "r") as include_file:
            include_file_data = include_file.read()

        include_cache[path] = include_file_data

    return include_cache[path]
